#!/usr/bin/perl -w

eval 'exec /usr/bin/perl -w -S $0 ${1+"$@"}'
    if 0; # not running under some shell
# $Id: tv_grab_fr,v 1.0 2003/01/31 23:04:20 epaepa Exp $

=head1 NAME

tv_grab_fr - Grab TV listings for France.

=head1 SYNOPSIS

tv_grab_fr [--help] [--output FILE] [--days N] [--offset N] [--quiet] [--proxy PROXY] [--longlisting]

=head1 DESCRIPTION

Output TV listings for several channels available in France.  The data
comes from tf1.guidetele.com.  The default is to grab as many days as
possible from the current day onwards. By default the program description are not
downloaded, so if you want description and ratings, you should active the --longlisting option.

B<--output FILE> write to FILE rather than standard output.

B<--days N> grab N days starting from today, rather than as many as
possible.

B<--offset N> start grabbing N days from today, rather than starting
today.  N may be negative.

B<--quiet> suppress the progress messages normally written to standard
error.

B<--longlisting> Gets additionnal information from the website, like program description and
ratings

B<--proxy PROXY> Use the given PROXY address instead of direct Internet access.

=head1 TODO

=item *

Add the support for the --configure option.

=item *

Add the support for french satellite TV (Canal+ and TPS).

=item *

Add the support for rare hertzien TV (RTL9, Suisse, ...).

=item *

Add the support for radios (?).


=head1 SEE ALSO

L<xmltv(5)>

=head1 AUTHOR

Sylvain Fabre, centraladmin@lahiette.com
based on a preliminary work from
Vincent Saugey, vsaugey@free.fr

=cut

# Todo: perhaps we should internationalize messages and docs?
use XMLTV::Usage <<END
$0: get France television listings in XMLTV format
usage: $0 tv_grab_fr [--help] [--output FILE] [--days N] [--offset N] [--quiet] [--proxy PROXY] [--longlisting]
END
  ;

# Use Term::ProgressBar if installed.
use constant Have_bar => eval { require Term::ProgressBar; 1 };

use strict;
use HTML::Entities;
use Getopt::Long;
use Date::Manip;
use LWP::Simple;
use LWP::UserAgent;
use XMLTV;
use XMLTV::Memoize;


#***************************************************************************
# Main declarations
#***************************************************************************
my $GRID_URL = 'http://tf1.guidetele.com/index_grille.html?b=16528';
my $SHEET_URL = "http://tf1.guidetele.com/fiche.html?id=";
my $MAX_STARS = 3;
my $HOURS_PER_PAGE = 4;
my $HOURS_PER_DAY = 24;
if ($HOURS_PER_DAY % $HOURS_PER_PAGE) {
    die "cannot yet handle pages which don\'t neatly divide a day";
}
my @offsets;
my %SummerWinterHours = ( "20040328020000" => "+0100",
                          "20041031030000" => "+0200",
                          "20050327020000" => "+0100",
                          "20051030030000" => "+0200",
                          "20060326020000" => "+0100",
                          "20061029030000" => "+0200" );

#***************************************************************************
# Sub sections
#***************************************************************************
sub GetTimeOffset {
    my ($swdate, $mydate) = $_[0];
    foreach $swdate (sort keys %SummerWinterHours) {
        if ( (&Date_Cmp($mydate, $swdate )) < 0 ) {
            return $SummerWinterHours{$swdate};
        }
    }
}

sub xmlencoding {
    # encode for xml
    $_[0] =~ s/</&lt;/g;
    $_[0] =~ s/>/&gt;/g;
    $_[0] =~ s/&/&amp;/g;
    return $_[0];
}

sub tidy {
    # clean bad characters from HTML
    for (my $s = shift) {
	tr/\205//d;
	tr/\222/''/;
	s/\234/oe/g;
	return $_;
    }
}

#***************************************************************************
# Options processing
#***************************************************************************
XMLTV::Memoize::check_argv('get_page');
my ($opt_days,  $opt_help,  $opt_output,  $opt_offset,  $opt_quiet,  $opt_configure, $opt_ll, $opt_proxy );
$opt_offset = 0;
$opt_quiet = 0;
$opt_days = 5;
$opt_output = '-'; # standard output
GetOptions('days=i'    => \$opt_days,
     'help'      => \$opt_help,
     'output=s'  => \$opt_output,
     'offset=i'  => \$opt_offset,
     'quiet'     => \$opt_quiet,
     'configure' => \$opt_configure,
     'longlisting' => \$opt_ll,
     'proxy=s' => \$opt_proxy
    )
  or usage(0);
die 'Number of days must not be negative'  if (defined $opt_days && $opt_days < 0);
usage(1) if $opt_help;
if ($opt_configure) {
    print STDERR "No configuration necessary : all the 6 TV channels are downloaded (TF1, A2, F3, C+, TV5/Arte, M6)\n";
    print STDERR "The next releases will integrate Satellite channels and maybe radio\n";
    exit();
}

#***************************************************************************
# Output the header of the file
#***************************************************************************
open OUT, ">$opt_output" or die "cannot write to $opt_output: $!";
print OUT <<END
<?xml version="1.0" encoding="ISO-8859-1"?>
<!DOCTYPE tv SYSTEM "xmltv.dtd">

<tv>
END
;

# Build the hours on a 24 to 23 basis, because guidetele manage dates in a very complicate and strange way
for (my $i = 0; $i < $HOURS_PER_DAY; $i += $HOURS_PER_PAGE) {
    if ($i == 0) {
        push @offsets, $HOURS_PER_DAY;
    } else {
        push @offsets, $i;
    }
}

my $bar = new Term::ProgressBar('Downloading grid', $opt_days * @offsets)
  if Have_bar and not $opt_quiet;
my ($first_day, $last_day) = ($opt_offset, $opt_offset + $opt_days - 1);
my %res = ();
my %channelList = ();
my %results;
my ($detailsCount , $gridCount, $lineCount) = 0;
my ($got, $channel);

# Set the user agent for page download
my $ua = LWP::UserAgent->new;
if ($opt_proxy) {
    $ua->proxy(['http', 'ftp'],  $opt_proxy);
}

# Get a page using this agent.
sub get_page( $ ) {
    my $url = shift;
    my $response = $ua->get($url);
    die $response->status_line if not $response->is_success;
    return $response->content;
}

#***************************************************************************
# Now process the days by getting the main grids.
#***************************************************************************
my ($flag, $date, $dateIndex, $dateUrl, $url, $response, $page);
# Now loop thru days
foreach my $day ($first_day..$last_day) {
    $date = &DateCalc("today","+$day days");
    $dateIndex = &UnixDate($date, "%Y%m%d");
    $dateUrl   = &UnixDate($date, "%Y-%m-%d");
    foreach (@offsets) {
        $url = "$GRID_URL" . "&h=$_&d=$dateUrl";
	$got = tidy(get_page($url));
	$res { $dateIndex } .= $got;
	$lineCount += ($got =~ s/\n/\n/mg);
	$gridCount++;
        update $bar if Have_bar and not $opt_quiet;
    }
}
# Out put some stats about pages
print STDERR "Number of grid : $gridCount\n" if not $opt_quiet;
my $bar_desc = new Term::ProgressBar('Getting descrip.', $lineCount)  if Have_bar and not $opt_quiet;

#***************************************************************************
# Now parse the grid an get specific program information
#***************************************************************************
my ($prog, $heureDebut, $heureFin, $star_rating, $programme, $timeoffset);
my ($startdate, $restype, $sval, $shour, $smin, $eval, $ehour, $emin, $enddate, $err, $tmpdate);
foreach (keys %res) {
    $date = $_;
    foreach (split '\n', $res{$_}) {
        update $bar_desc if Have_bar and not $opt_quiet;
        # Detect channel name/id
        if (/<img alt=\"(.*)\" border=\"\d*\"/) {
            $channel = $1;
            $channel =~ s/\s//g;
            if ( !$channelList{ $channel } ) {
                $channelList{ $channel } = "NOLOGO";
            }
        }
        # Now detect logos
        if (/src=\"\/img\/chaine\/(.*)\"/) {
            $channelList{ $channel } = $1;
        }
        if (/onmouseover/) {
            next if (!m/onmouseover="drc\(([^"]+)\)" onmouseout=/);  # " To avoid dummy syntax color in some editors
            $1 =~ m/\'(.*)\',\'(.*)\'/;
            $prog->{name} = $2;
            next if ( $prog->{name} eq "Fin des programmes");
            ($prog->{heure}, $prog->{type}, $prog->{showview}) = split (/<br>&nbsp;/,$1);
            next if ( !$prog->{heure} );
            ($heureDebut, $heureFin)  = split("-", $prog->{heure});
            # Sometimes the type is nt set, so replace it by the showview field
            if (defined $prog->{type} and $prog->{type} =~ m/Showview : /) {
                $prog->{showview} = $prog->{type};
                undef $prog->{type};
            }
            # Process the type if defined : the type is sometimes defined with "TYPE, SUBTYPE - DATE"
            if  (defined $prog->{type} ) {
                ($prog->{type}, $prog->{datecreate}) = split("&nbsp;-&nbsp;", $prog->{type});
                ($prog->{type}, $prog->{subtype}) = split(", ", $prog->{type});
                $prog->{type} = xmlencoding(decode_entities( ucfirst($prog->{type}) ) );
                $prog->{subtype} = xmlencoding(decode_entities( ucfirst($prog->{subtype}) ) ) if defined $prog->{subtype};
            }
            # Process the showview field
            $prog->{showview} =~ s/Showview : // if defined $prog->{showview};
            # Process the name field
            $prog->{name} =~ s/^\d{7} //;
            $prog->{name} =~ s/\\//g;
            $prog->{name} = xmlencoding( decode_entities( $prog->{name} ) );

            if (($heureDebut =~ s/h//g) == 1) {
                ($sval, $shour, $smin) = split(/([0-9][0-9])([0-9][0-9])/, $heureDebut);
                ($eval, $ehour, $emin) = split(/([0-9][0-9])h([0-9][0-9])/, $heureFin);
                # Parse/split the date for processing
                my ($olddate, $myyear, $mymonth, $myday) = split(/([0-9][0-9][0-9][0-9])([0-9][0-9])([0-9][0-9])/, $date );
                $tmpdate = &ParseDateString($myyear.$mymonth.$myday);
                # This is very strange, but guidetele.com mismatch hours 0-4 and date :
                #    when hours are between 0 and 4, then this is the program for the previous day.
                if ( $shour < 4 && $ehour < 4) {
                    $startdate=&UnixDate(&DateCalc($tmpdate,"+1 day") , '%Y%m%d' );
                    $enddate=&UnixDate(&DateCalc($tmpdate,"+1 day") , '%Y%m%d' );
                } else {
                    $startdate = $date;
                    $enddate = $date;
                }
                # In this case, we are in a day frontier (program start before 00:00 and ends at 00, 01, ...)
                # The website does not change the date properly, so take care of this
                if ( ($ehour - $shour) < 0 ) {
                    $enddate=&UnixDate(&DateCalc($tmpdate,"+1 day") , '%Y%m%d' );
                }
                $timeoffset = GetTimeOffset($startdate.${heureDebut}."00");
                $programme = "  <programme start=\"$startdate${heureDebut}00 $timeoffset\"";
                $programme .=  " stop=\"$enddate${heureFin}00 $timeoffset\"" if (($heureFin =~ s/h//g) == 1);
                $programme .=  " showview=\"".$prog->{showview}."\"" if defined $prog->{showview};
                $programme .=  " channel=\"$channel\"";

                if ($prog->{name} =~ s/\s+([*]+)\s*$//) {
                    my $n = length $1;
                    if (0 < $n and $n <= $MAX_STARS) {
                        $star_rating = "$n/$MAX_STARS";
                    } elsif ($MAX_STARS < $n) {
                        warn "too many stars ($n), expected at most $MAX_STARS\n";
                    } else { die }
                }
                warn "seemingly two star-ratings for a programme\n" if $prog->{name} =~ /[*]$/;
                $programme .= ">\n    <title>".$prog->{name}."</title>\n";
                # Now get program description if the longlisting option is set
                if ( $opt_ll && m/href=\'javascript\:of\((\d+)\)/ ) { # ' To avoid dummy syntax color in some editors
                    $page = tidy(get_page($SHEET_URL . $1));
                    $detailsCount++;
                    if ( $page =~ m/<B>RESUME<\/B><br>([^\n]+)/) {
                        my $desc = xmlencoding(decode_entities($1));
                        $desc =~ s/\s+$//;
                        $programme .= "    <desc lang=\"fr\">$desc</desc>\n" if length $desc;
                    }
                }
                $programme .= "    <category>".$prog->{type}."</category>\n" if defined $prog->{type};
                $programme .= "    <category>".$prog->{subtype}."</category>\n" if defined $prog->{subtype};
                $programme .= "    <date>".$prog->{datecreate}."</date>\n" if defined $prog->{datecreate};
                if (defined $star_rating) {
                    $programme .= "    <star-rating>\n";
                    $programme .= "      <value>$star_rating</value>\n";
                    $programme .= "    </star-rating>\n";
                }
                $programme .= "  </programme>\n";
                # We must set the key to start+channel to avoid duplicated start values
                $results{$startdate.$heureDebut."00".$channel} = $programme;
            }
        }
    }
}

print STDERR "Number of details pages : $detailsCount\n" if not $opt_quiet and $opt_ll;

#***************************************************************************
# Now output the results in the file
#***************************************************************************
# Now output the channel list
foreach (keys %channelList ) {
    print OUT <<END
  <channel id="$_">
    <display-name lang="fr">$_</display-name>
END
  ;
    for ($channelList{$_}) {
        print OUT qq'    <icon src="http://tf1.guidetele.com/img/chaine/$_" />\n' unless /NOLOGO/;
    }
    print OUT "  </channel>\n";
}

# Now output the program grid itself, sorted by date+time
foreach ( sort keys %results ) {
    for ($results{$_}) {
        print OUT $_;
    }
}


print OUT "</tv>\n";
